using namespace std;
#include <iostream>
#include <cstring>

/*
 * ラムダ計算をするプログラム
 *
 * こんな感じの式が評価できます
 * (Lambda x.(x x) Lambda x.(x x))
 *
 * TODO:
 * 「Lambda (x x).x」のような
 * 変なラムダ式も、評価の結果作れてしまいます
 */

// ------------------------------------------------------------------------
// Expression := <Lambda>|<Pair>|<Var>
class Var;
class Expression{
public:
  virtual void print( void ){ cerr << "error\n";  }
  virtual Expression *Eval(void){ return nullptr; }
  virtual Expression *Beta( Expression *x ){ return nullptr; }
  virtual Expression *dup(void) = 0;
  virtual Expression *replace( Expression *n, Expression *x ) = 0;
  virtual bool comp_n( Expression *x ){ return false; }
  virtual bool comp_n( Var *x ){ return false; }
};
// Pair := (<Expression> <Expression>)
class Pair:public Expression{
  Expression *first;
  Expression *second;
public:
  Pair( Expression *f, Expression * s ){
    first = f;
    second = s;
  }
  void print( void ){
    std::cout << "(";
    first->print(); 
    std::cout << " ";
    second->print();
    std::cout << ")";
  }
  Expression * Eval( void ){
    auto r = first->Beta( second );
    if( r != nullptr ) return r;
    r = first->Eval();
    if( r != nullptr ) return new Pair( r, second->dup() );
    r = second->Eval();
    if( r != nullptr ) return new Pair( first->dup(), r );
    return nullptr;
  }
  Expression * replace( Expression*n, Expression *x ){
    return new Pair( first->replace(n,x), second->replace(n,x) );
  }
  Expression * dup( void ){
    return new Pair(first->dup(),second->dup());
  }
  ~Pair( void ){
    delete first;
    delete second;
  }
};
// Var := [a-zA-Z]+
class Var: public Expression{
  string name;
public:
  Var( string n ){
    name = n;
  }
  void print( void ){
    std::cout << name;
  }
  Expression * dup( void ){
    return new Var( name );
  }
  Expression * replace( Expression*n, Expression *x ){
    if( n->comp_n( this )  ){
      return x->dup();
    }else{
      return this->dup();
    }
  }
  bool comp_n( Var *n ){
    return name == n->name;
  }
};

// Lambda := Lambda <Expression>.<Expression>
class Lambda: public Expression {
  Expression *var;
  Expression *exp;
public:
  Lambda( Expression *v, Expression *x ){
    var = v;
    exp = x;
  }
  void print( void ){
    std::cout << "Lambda ";
    var->print();
    std::cout << ".";
    exp->print();
  }
  Expression * replace( Expression *v, Expression *x ){
    return new Lambda( var->replace( v, x )  ,exp->replace( v, x ) );
  }
  Expression * dup( void ){
    return new Lambda( var->dup(), exp->dup() );
  }
  Expression * Eval( void ){
    auto r = exp->Eval();
    if( r != nullptr)
      return new Lambda( var->dup(), r );
    return nullptr;
  }
  Expression *Beta( Expression *x ){
    return exp->replace( var, x );
  }
  ~Lambda(void){
    delete var;
    delete exp;
  }
};

// --------------------------------------------------------------------------
string buf;
int    nextChr;

int getNextChr( void ){
  nextChr = cin.get();
  return nextChr;
}

#define SPC_CHARS (" \n\t\r")
inline bool isSPC( int c ){
  return strchr( SPC_CHARS, c );
}

enum _TOKEN{
  T_NONE, T_LBR,T_RBR,T_DOT,T_LAMBDA,T_VAR
};

_TOKEN getToken(){ 
  while( isSPC(nextChr)  ){
    getNextChr();
  }
  if( nextChr == '(' ){
    getNextChr();
    return T_LBR;
  } else if( nextChr == ')' ){
    getNextChr();
    return T_RBR;
  } else if( nextChr == '.' ){
    getNextChr();
    return T_DOT;
  } else if( isalpha(nextChr) ) {
    buf = "";
    do{
      buf += nextChr;
      getNextChr();
    }  while( isalpha(nextChr) );
    if( buf  == "Lambda" ) return T_LAMBDA;
    return T_VAR;
  }
  return T_NONE;
}

// --------------------------------------------------------------------------
Expression *getExpression(void){
  _TOKEN token = getToken();
  if( token == T_LAMBDA ){
    if( getToken() != T_VAR ){
      cerr << "ERROR: token is expected.\n";
      return nullptr;
    }
    auto var = new Var(buf);
    if( getToken() != T_DOT){
      cerr << "ERROR: . is expected.\n";
      return nullptr;
    }
    Expression *exp = getExpression();
    if( exp == nullptr ){
      cerr << "ERROR: expression is expected.\n";
      return nullptr;
    }
    return new Lambda(var,exp);
  } else if( token == T_LBR ){
    Expression *first = getExpression(); 
    if( first == nullptr){
      cerr << "ERROR: first expression is expected.\n";
      return nullptr;
    }
    Expression *second = getExpression();
    if(second == nullptr){
      delete first;
      cerr << "ERROR: second expression is expected.\n";
      return nullptr;
    }
    if( getToken() != T_RBR ){
      delete first;
      delete second;
      cerr << "ERROR: ) is expected.\n";
      return nullptr;
    }
    return new Pair( first, second );
  }else if(token == T_VAR){
    return new Var(buf);
  } 
  cerr << "ERROR: unexpected EOF.\n";
  return nullptr;
}

// --------------------------------------------------------------------------
main(void){
  cerr << "Input lambda expression.>\n";

  getNextChr();
  Expression *exp = getExpression();
  cout << "--- Evaluate --\n";

  while(exp != nullptr){
    exp->print();
    cout << "\n-> ";
    auto nextExp = exp->Eval();
    delete exp;
    exp = nextExp;
  }
  cout << "Quit.\n";
}

//EOF
